/*
 * QEMU PC System Emulator
 *
 * Copyright (c) 2003-2004 Fabrice Bellard
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
/*
 *  Q35 chipset based pc system emulator
 *
 *  Copyright (c) 2009, 2010
 *                     Isaku Yamahata <yamahata at valinux co jp>
 *                     VA Linux Systems Japan K.K.
 *
 *  This is based on pc.c, but heavily modified.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, see <http://www.gnu.org/licenses/>
 */
#include "hw.h"
#include "pc.h"
#include "fdc.h"
#include "pci.h"
#include "pci_bridge.h"
#include "i21154.h"
#include "pcie_root.h"
#include "pcie_upstream.h"
#include "pcie_downstream.h"
#include "pcie_msix_template.h"
#include "block.h"
#include "sysemu.h"
#include "audio/audio.h"
#include "net.h"
#include "smbus.h"
#include "boards.h"
#include "monitor.h"
#include "fw_cfg.h"
#include "hpet_emul.h"
#include "watchdog.h"
#include "smbios.h"
#include "ide.h"
#include "usb-uhci.h"

#include "q35.h"

#define MAX_IDE_BUS 2

static const int ide_iobase[MAX_IDE_BUS] = { 0x1f0, 0x170 };
static const int ide_iobase2[MAX_IDE_BUS] = { 0x3f6, 0x376 };
static const int ide_irq[MAX_IDE_BUS] = { 14, 15 };

static void pc_q35_bridge_init(PCIDevice *gmch_state,
                               PCIBus *host_bus, PCIBus *pci_bus)
{
    uint8_t dev;
    uint8_t sec_bus;
    uint8_t sub_bus;
    uint8_t port = 0;
    uint8_t chassis = 0;
    uint16_t slot = 0;
    PCIESlot *s;
    uint8_t fn;
    PCIESlot *root_port;
    PCIBus *root_port_bus;
    char buf[16];

    /* PCI to PCI bridge b6:d[29 - 31]:f0, 6:[1c - 1f].0 with subordinate bus
       of 7 - 9 on b0:d30:f0, 0.1e.0 = bus */
#define Q35_P2P_BRDIGE_DEV_BASE         28
#define Q35_P2P_BRDIGE_DEV_MAX          32
#define Q35_P2P_BRDIGE_SUBBUS_BASE      (ICH9_D2P_SECONDARY_DEFAULT + 1)
    for (dev = Q35_P2P_BRDIGE_DEV_BASE; dev < Q35_P2P_BRDIGE_DEV_MAX; dev++) {
        PCIBridge *br;
        sec_bus = Q35_P2P_BRDIGE_SUBBUS_BASE + dev - Q35_P2P_BRDIGE_DEV_BASE;
        sub_bus = sec_bus;

        snprintf(buf, sizeof(buf), "pci.%d", sec_bus);
        br = i21154_init(pci_bus, PCI_DEVFN(dev, 0), buf, sec_bus, sub_bus);

        /* pci hotplug */
        gmch_bus_hotplug(gmch_state, pci_bridge_get_sec_bus(br));
    }

    /* PCIe root port b0:d4:f0, 0:04.0 with subordinate bus of 63 */
    sec_bus = 32;
    sub_bus = sec_bus;
    snprintf(buf, sizeof(buf), "pcie.%d", sec_bus);
    s = pcie_root_init(host_bus, PCI_DEVFN(4, 0),
                       buf, sec_bus, sub_bus, port, chassis, slot);

#define Q35_MANY_SLOTS
//#undef Q35_MANY_SLOTS

#ifdef Q35_MANY_SLOTS
#define Q35_NR_ROOTPORT         6
#define Q35_NR_UPSTREAM         8
#define Q35_NR_DOWNSTREAM       16
#else
#define Q35_NR_ROOTPORT         1
#define Q35_NR_UPSTREAM         1
#define Q35_NR_DOWNSTREAM       1
#endif

    /* PCIe root port b0:d24:f[0-5], 0.18.[0-5] */
    for (fn = 0; fn < Q35_NR_ROOTPORT; fn++) {
        sec_bus++;
        sub_bus = sec_bus;
        port++;
        slot++;

        snprintf(buf, sizeof(buf), "pcie.%d", sec_bus);
        s = pcie_root_init(host_bus, PCI_DEVFN(24, fn),
                           buf, sec_bus, sub_bus, port, chassis, slot);
    }

    /* PCIe root port b0:d25:f0 */
    sub_bus = ICH9_D2P_SECONDARY_DEFAULT - 1;
    sec_bus = sub_bus - (1 + 8 + 8 * 16);
    port++;
    slot++;
    snprintf(buf, sizeof(buf), "pcie.%d", sec_bus);
    root_port = pcie_root_init(host_bus, PCI_DEVFN(25, 0),
                               buf, sec_bus, sub_bus, port, chassis, slot);
    root_port_bus = pci_bridge_get_sec_bus(&root_port->port.br);

    /* 8 * 16 = 128 slots */
    port = 0;
    for (fn = 0; fn < Q35_NR_UPSTREAM; fn++) {
        PCIEPort *upstream;
        PCIBus *upstream_bus;
        uint16_t downstream_port;

        /* PCIe upstream port d0:f[0-7] */
        sec_bus++;
        sub_bus = sec_bus + 16;
        snprintf(buf, sizeof(buf), "pcie.%d", sec_bus);
        upstream = pcie_upstream_init(root_port_bus, PCI_DEVFN(0, fn),
                                      buf, sec_bus, sub_bus, port);
        upstream_bus = pci_bridge_get_sec_bus(&upstream->br);
        port++;

        /* PCIe downstream port */
        downstream_port = 0;
        for (dev = 0; dev < Q35_NR_DOWNSTREAM; dev++) {
            sec_bus++;
            sub_bus = sec_bus;
            slot++;
            snprintf(buf, sizeof(buf), "pcie.%d", sec_bus);
            pcie_downstream_init(upstream_bus, PCI_DEVFN(dev, 0),
                                 buf, sec_bus, sub_bus, downstream_port,
                                 chassis, slot);
            downstream_port++;
        }
    }

    /* initialize the pcie_msix device template */
    pcie_msix_init(root_port_bus);
}

/* PC hardware initialisation */
static void pc_q35_init(ram_addr_t ram_size,
                        const char *boot_device,
                        const char *kernel_filename,
                        const char *kernel_cmdline,
                        const char *initrd_filename,
                        const char *cpu_model)
{
    int i;
    ram_addr_t below_4g_mem_size, above_4g_mem_size;
    FWCfgState *fw_cfg;
    PCIBus *host_bus;
    PCIBus *pci_bus;
    static PCIDevice *gmch_state;
    int ich9_lpc_devfn = -1;
    qemu_irq *cpu_irq;
    qemu_irq *isa_irq;
    qemu_irq *i8259;
    IsaIrqState *isa_irq_state;
    DriveInfo *hd[MAX_IDE_BUS * MAX_IDE_DEVS];
    FDCtrl *floppy_controller;
    ISADevice *rtc_state;
    PCIDevice *lpc;
    qemu_irq *cmos_s3;

    pc_cpus_init(cpu_model);

    vmport_init();

    /* allocate ram and load rom/bios */
    fw_cfg = pc_memory_init(ram_size, kernel_filename, kernel_cmdline,
                            initrd_filename,
                            &below_4g_mem_size, &above_4g_mem_size);

    cpu_irq = pc_allocate_cpu_irq();
    i8259 = i8259_init(cpu_irq[0]);
    isa_irq_state = qemu_mallocz(sizeof(*isa_irq_state));
    isa_irq_state->i8259 = i8259;
    isa_irq = qemu_allocate_irqs(isa_irq_handler, isa_irq_state, 24);

    host_bus = gmch_init(&gmch_state, isa_irq, &ich9_lpc_devfn,
                         &isa_irq_state->ioapic);

    /* create ISA bus */
    lpc = gmch_lpc_init(host_bus, gmch_state, &ich9_lpc_devfn);
    isa_bus_irqs(isa_irq);
    pc_register_ferr_irq(isa_reserve_irq(13));

    pci_bus = ich9_d2pbr_init(host_bus, PCI_DEVFN(ICH9_D2P_BRIDGE_DEV,
                                                  ICH9_D2P_BRIDGE_FUNC),
                              ICH9_D2P_SECONDARY_DEFAULT);
    pci_set_default_bus(pci_bus);

    pc_vga_init(host_bus, fw_cfg);

    /* init basic PC hardware */
    pc_basic_device_init(isa_irq, &floppy_controller, &rtc_state);

    /* connect lpc pm stuff */
    cmos_s3 = qemu_allocate_irqs(pc_cmos_set_s3_resume, rtc_state, 1);
    ich9_lpc_pm_init(lpc, host_bus, gmch_state, *cmos_s3);

    pc_q35_bridge_init(gmch_state, host_bus, pci_bus);

    for(i = 0; i < nb_nics; i++) {
        NICInfo *nd = &nd_table[i];

        pci_nic_init(nd, "e1000", NULL);
    }

    if (drive_get_max_bus(IF_IDE) >= MAX_IDE_BUS) {
        fprintf(stderr, "qemu: too many IDE bus\n");
        exit(1);
    }

    for(i = 0; i < MAX_IDE_BUS * MAX_IDE_DEVS; i++) {
        hd[i] = drive_get(IF_IDE, i / MAX_IDE_DEVS, i % MAX_IDE_DEVS);
    }

    /* XXX ich9 supports only SATA */
    pci_piix3_ide_init(pci_bus, hd, -1);

#ifdef HAS_AUDIO
    pc_audio_init(pci_bus, isa_irq);
#endif

    pc_cmos_init(below_4g_mem_size, above_4g_mem_size, boot_device, hd,
                 floppy_controller, rtc_state);

    if (usb_enabled) {
        /* XXX: make ich9 specific add usb_uchi_ich9_init() */
        usb_uhci_piix4_init(pci_bus, -1);
    }

    if (acpi_enabled) {
        uint8_t *eeprom_buf = qemu_mallocz(8 * 256); /* XXX: make this persistent */
        i2c_bus *smbus;

        /* TODO: Populate SPD eeprom data.  */
        /* XXX determine proper smb_io_base */
        smbus = ich9_smb_init(host_bus, ich9_lpc_devfn + 3, 0xb100);
        for (i = 0; i < 8; i++) {
            DeviceState *eeprom;
            eeprom = qdev_create((BusState *)smbus, "smbus-eeprom");
            qdev_prop_set_uint8(eeprom, "address", 0x50 + i);
            qdev_prop_set_ptr(eeprom, "data", eeprom_buf + (i * 256));
            qdev_init_nofail(eeprom);
        }
    }

    if (gmch_state) {
        gmch_init_memory_mappings(gmch_state);
    }

    pc_pci_device_init(pci_bus);
    pc_pci_bus_number_add(fw_cfg);
}

static QEMUMachine pc_q35_machine = {
    .name = "pc_q35",
    .desc = "Q35 chipset PC",
    .init = pc_q35_init,
    .max_cpus = 255,
};

static void pc_q35_machine_init(void)
{
    qemu_register_machine(&pc_q35_machine);
}

machine_init(pc_q35_machine_init);
