// LICENSETEXT
//
//   Copyright (C) 2008 : GreenSocs Ltd
//       http://www.greensocs.com/ , email: info@greensocs.com
//
//   Developed by :
//
//   Màrius Montón <marius.monton@uab.cat>,
//     Hardware-Software Prototypes and Solutions Lab (CEPHIS).
//     Microelectronics and Electronics SystemS Dpt. UAB.
//     http://cephis.uab.cat
//
//   This program is free software.
//
//   If you have no applicable agreement with GreenSocs Ltd, this software
//   is licensed to you, and you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   If you have a applicable agreement with GreenSocs Ltd, the terms of that
//   agreement prevail.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
//   02110-1301  USA
//
// ENDLICENSETEXT

/*
 * PCI device emulation
 */

#include "hw.h"
#include "pc.h"
#include "pci.h"
#include "qemu-timer.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>

/* debug SCLink code */
#define DEBUG_SCLINK_PCI

#ifdef DEBUG_SCLINK_PCI
#define PRINT_DEBUG(format, ...) printf(format, __VA_ARGS__);
#define PRINT_DEBUG0(format) printf(format);
#else
#define PRINT_DEBUG(format, ...)
#define PRINT_DEBUG0(format)
#endif

/* Common declarations for all PCI devices. */

#define PCI_CONFIG_8(offset, value) (pci_conf[offset] = (value))
#define PCI_CONFIG_16(offset, value) (*(uint16_t *)&pci_conf[offset] = cpu_to_le16(value))
#define PCI_CONFIG_32(offset, value) (*(uint32_t *)&pci_conf[offset] = cpu_to_le32(value))
    
typedef uint32_t (*sc_read_T)(uint32_t pos, void* instPtr, uint32_t len);
typedef void (*sc_write_T)(uint32_t pos, uint32_t val, void* instPtr, uint32_t len);
typedef uint32_t (*sc_read_dmi_T)(uint32_t pos, void* instPtr, uint32_t len);
typedef void (*sc_write_dmi_T)(uint32_t pos, uint32_t val, void* insPtr, uint32_t len);

extern sc_read_T sc_read;
extern sc_write_T sc_write;
extern sc_read_dmi_T sc_read_dmi;
extern sc_write_dmi_T sc_write_dmi;

extern void* ourWrapper;

/* Configuration registers */
#define SC_LINK_VEN_ID      0x1172
#define SC_LINK_DEV_ID      0x0004
#define SC_LINK_REV_ID      0x01
#define SC_LINK_CLASS       0xFA0000	/* Warning: Not use 0xFF0000 */

#define SC_LINK_CIS_PTR     0x00000000	/* CardBus CIS pointer */

#define SC_LINK_SUB_VEN_ID  0x0000 	    /* Subsystem vendor ID*/
#define SC_LINK_SUB_ID      0x0000		/* Subsytem ID */
#define SC_LINK_CAP_PTR     0x40		/* Capabilities list pointer */
#define SC_LINK_INT_PIN     0x01		/* Interrupt pin register 0-> No IRQ; 1-> IRQ used */
#define SC_LINK_MIN_GNT     0x00		/* Minimum grant */
#define SC_LINK_MAX_LAT     0x00		/* Maximum latency */

/* PCI BUS commands */
#define CMD_IOREAD	        0x02
#define CMD_IOWRITE	        0x03
#define CMD_MEMREAD	        0x06
#define CMD_MEMWRITE	    0x07
#define CMD_CONFIGREAD	    0x0A
#define CMD_CONFIGWRITE	    0x0B

static const int BAR_Regions[6][2] = 
{
	// len , type 
	{ 0x200, PCI_BASE_ADDRESS_SPACE_MEMORY} ,   //BAR0
	{ 0x200, PCI_BASE_ADDRESS_SPACE_IO} ,       //BAR1
	{ 0, 0} ,                                   //BAR2
	{ 0, 0} ,                                   //BAR3
	{ 0, 0} ,                                   //BAR4
	{ 0, 0}                                     //BAR5 
};

#if 0
static const int BAR_Regions_test[6][2] = 
{
	// len , type 
	{ 0x400, PCI_BASE_ADDRESS_SPACE_IO} ,       //BAR0
	{ 0x100, PCI_BASE_ADDRESS_SPACE_MEMORY} ,   //BAR1
	{ 0x800, PCI_BASE_ADDRESS_SPACE_IO} ,       //BAR2
	{ 0, 0} ,                                   //BAR3
	{ 0, 0} ,                                   //BAR4
	{ 0, 0}                                     //BAR5 
};
#endif 

typedef struct SCLinkState_st 
{
    PCIDevice dev;
    int mmio_index;
} SCLinkState;

qemu_irq sc_link_pci_interrupt;

void sc_raise_pci_irq();
void sc_low_pci_irq();
void sc_link_pci_init(PCIBus *bus);

static void 
sc_link_configuration_write(uint8_t *pci_conf)
{
    *(uint16_t *)pci_conf = cpu_to_le16(SC_LINK_VEN_ID);        // Vendor ID
    *(uint16_t *)(pci_conf + 2) = cpu_to_le16(SC_LINK_DEV_ID);  // Device ID

    pci_conf[0x08] = SC_LINK_REV_ID;
    pci_conf[0x09] = (uint8_t) (SC_LINK_CLASS & 0xff);          // Class Code
    pci_conf[0x0a] = (uint8_t) (SC_LINK_CLASS >> 8); 
    pci_conf[0x0b] = (uint8_t) (SC_LINK_CLASS >> 16); 

    pci_conf[0x2c] = (uint8_t) (SC_LINK_SUB_VEN_ID & 0xff);     //Subsystem vendor id
    pci_conf[0x2d] = (uint8_t) (SC_LINK_SUB_VEN_ID >> 8);
    pci_conf[0x2e] = (uint8_t) (SC_LINK_SUB_ID & 0xff);	        //subsystem id
    pci_conf[0x2f] = (uint8_t) (SC_LINK_SUB_ID >> 8);

    pci_conf[0x34] = SC_LINK_CAP_PTR;
    //pci_conf[0x3c] = SC_LINK_PIN_LN;    
    pci_conf[0x3d] = SC_LINK_INT_PIN; 
    pci_conf[0x3e] = SC_LINK_MIN_GNT;
    pci_conf[0x3f] = SC_LINK_MAX_LAT;
}

static void 
sc_link_pci_io_write(PCIDevice *d, uint32_t addr, uint32_t val, int len)
{
    uint32_t reg_base, reg_size;
    int bar_region=0, i;
    
    for(i=0; i<PCI_NUM_REGIONS; i++) 
    {
	    reg_base = d->io_regions[i].addr;

	    if (reg_base < 0) 
            continue; /*-1 means not mapped */
    	
	    reg_size = d->io_regions[i].size;
	    if((reg_base <= addr) && (reg_base+reg_size > addr)) 
	    {
	        bar_region = i;
	        break;
	    }
    }

    PRINT_DEBUG("\n%s: reg_base: %08X\n", __FUNCTION__, reg_base);
    PRINT_DEBUG("\n%s: io_write: reg_base=0x%x len=0x%x addr=0x%x val=0x%x bar=%d\n", __FUNCTION__, reg_base, len, addr, val, bar_region);

    sc_write(addr-reg_base, val, ourWrapper, 1);
}

static uint32_t
sc_link_pci_io_read(PCIDevice *d, uint32_t addr, int len)
{
    uint32_t reg_base, reg_size, val;
    int bar_region=0, i;
    
    for(i=0; i<PCI_NUM_REGIONS; i++) 
    {
	    reg_base = d->io_regions[i].addr;
	    if (reg_base < 0) 
            continue; /*-1 means not mapped */
    	
	    reg_size = d->io_regions[i].size;
	    if((reg_base <= addr) && (reg_base+reg_size > addr))
	    {
	        bar_region = i;
	        break;
	    }
    }
    
    PRINT_DEBUG("\n%s: io_read %d bytes. Addr=0x%x bar=%d",__FUNCTION__, len, addr, bar_region);

    val = sc_read(addr-reg_base, ourWrapper, 1);
    PRINT_DEBUG("\n%s: io_read %d bytes. Addr=0x%x val=0x%x bar=%d",__FUNCTION__, len, addr, val, bar_region);

    return val;
}

static void
sc_link_pci_io_writeb(void* opaque, uint32_t addr, uint32_t val)
{
    sc_link_pci_io_write(opaque, addr, val, 1);
}

static void
sc_link_pci_io_writew(void* opaque, uint32_t addr, uint32_t val)
{
    sc_link_pci_io_write(opaque, addr, val, 2);
}

static void
sc_link_pci_io_writel(void* opaque, uint32_t addr, uint32_t val)
{
    sc_link_pci_io_write(opaque, addr, val, 4);
}

static uint32_t
sc_link_pci_io_readb(void* opaque, uint32_t addr)
{
    return sc_link_pci_io_read(opaque, addr, 1);
}

static uint32_t
sc_link_pci_io_readw(void* opaque, uint32_t addr)
{
    return sc_link_pci_io_read(opaque, addr, 2);
}

static uint32_t
sc_link_pci_io_readl(void* opaque, uint32_t addr)
{
    return sc_link_pci_io_read(opaque, addr, 4);
}

static void 
sc_link_pci_io_map(PCIDevice *pci_dev, int region_num, pcibus_t addr, 
		   pcibus_t size, int type)
{
    register_ioport_write(addr, 16, 1, sc_link_pci_io_writeb, pci_dev);
    register_ioport_write(addr, 16, 2, sc_link_pci_io_writew, pci_dev);
    register_ioport_write(addr, 16, 4, sc_link_pci_io_writel, pci_dev);
    register_ioport_read(addr , 16, 1, sc_link_pci_io_readb, pci_dev);
    register_ioport_read(addr , 16, 2, sc_link_pci_io_readw, pci_dev);
    register_ioport_read(addr , 16, 4, sc_link_pci_io_readl, pci_dev);
}

static void
sc_link_pci_mem_writeb(void *opaque, target_phys_addr_t addr, uint32_t val)
{
    PRINT_DEBUG0("\nSC_LINK Core only 4Bytes transfers (writeb)!!!\n");
    //PCI Core only have 4Bytes transfers
}

static void
sc_link_pci_mem_writew(void *opaque, target_phys_addr_t addr, uint32_t val)
{
    PRINT_DEBUG0("\nSC_LINK Core only 4Bytes transfers (writew)!!!\n");
    //PCI Core only have 4Bytes transfers
}

static void 
sc_link_pci_mem_writel(void *opaque, target_phys_addr_t addr, uint32_t val)
{
    uint32_t reg_base, reg_size;
    uint32_t base_addr;
    int bar_region=0, i;
    PCIDevice *d=opaque;
    
    PRINT_DEBUG("\n%s: mem_write: addr: 0x%x\n", __FUNCTION__, addr);
    
    for(i=0; i<PCI_NUM_REGIONS; i++) 
    {
	    reg_base = d->io_regions[i].addr;
	    if (reg_base < 0) 
            continue; /*-1 means not mapped */
    	
	    reg_size = d->io_regions[i].size;
	    if((reg_base <= addr) && (reg_base+reg_size > addr)) 
        {
	        bar_region = i;
	        break;
	    }
    }
    base_addr = d->io_regions[bar_region].addr;
    PRINT_DEBUG("\n%s: mem_write BAR Base address=0x%x Addr=0x%x val=0x%x bar=%d\n", __FUNCTION__, base_addr, addr, val,bar_region);
    PRINT_DEBUG("\n%s: mem_write REAL address: 0x%x\n", __FUNCTION__, base_addr + addr);
    
    PRINT_DEBUG("%s: PCIIOREGIONS:\n", __FUNCTION__);
    
    for(i=0;i< PCI_NUM_REGIONS-1;i++) 
    {
	    PRINT_DEBUG("\t%d Addr: 0x%"FMT_PCIBUS" Size: 0x%"FMT_PCIBUS". Type: 0x%x\n", i, d->io_regions[i].addr, d->io_regions[i].size, d->io_regions[i].type);
    }
    
    // This BAR is memory mapped, we perform a DMI access instead of a standard access. 
    sc_write_dmi(addr-reg_base, val, ourWrapper, 1);
}


static uint32_t 
sc_link_pci_mem_readb(void *opaque, target_phys_addr_t addr)
{
    PRINT_DEBUG0("\nSC_LINK Core only 4Bytes transfers (readb)!!!\n");
    //PCI Core only have 4Bytes transfers	
    return 0x00000000;
}

static uint32_t 
sc_link_pci_mem_readw(void *opaque, target_phys_addr_t addr)
{
    PRINT_DEBUG0("\nSC_LINK Core only 4Bytes transfers (readw)!!!\n");
    //PCI Core only have 4Bytes transfers	
    return 0x00000000;
}

static uint32_t 
sc_link_pci_mem_readl(void *opaque, target_phys_addr_t addr)
{
    uint32_t val, reg_base, reg_size;
    int bar_region, i;
    PCIDevice *d=opaque;

    PRINT_DEBUG("\n%s: device name: %s\t devfn: %d\n", __FUNCTION__, d->name, d->devfn);
  
    bar_region = 0;
    for(i=0; i<PCI_NUM_REGIONS; i++) 
    {
	    reg_base = d->io_regions[i].addr;
	    if (reg_base < 0) continue; 
        
	    reg_size = d->io_regions[i].size;
	    if((reg_base <= addr) && (reg_base+reg_size > addr)) 
	    {
	        bar_region = i;
	        break;
	    }
    }

    PRINT_DEBUG("\n%s: device name: %s\t devfn: %d\n", __FUNCTION__, d->name, d->devfn);
    PRINT_DEBUG("\n%s: mem_read Addr=0x%x bar=%d\n", __FUNCTION__, addr-reg_base, bar_region);

    // This BAR is memory mapped, we perform a DMI access instead of a standard access. 
    val = sc_read_dmi(addr-reg_base, ourWrapper, 1);

    PRINT_DEBUG("\n%s: mem_read Addr=0x%x val=0x%x bar=%d\n", __FUNCTION__, addr-reg_base, val, bar_region);

    return val;
}

static CPUWriteMemoryFunc * const sc_link_pci_mem_write[] = {
    sc_link_pci_mem_writeb, sc_link_pci_mem_writew, sc_link_pci_mem_writel
};

static CPUReadMemoryFunc * const sc_link_pci_mem_read[] = {
    sc_link_pci_mem_readb, sc_link_pci_mem_readw, sc_link_pci_mem_readl
};

static void 
sc_link_pci_mem_map(PCIDevice *pci_dev, int region_num, 
		     pcibus_t addr, pcibus_t size, int type)
{
    SCLinkState*d = DO_UPCAST(SCLinkState, dev, pci_dev);
    
    cpu_register_physical_memory(addr, size, d->mmio_index);
}

void 
sc_raise_pci_irq()
{
    PRINT_DEBUG("\n%s: raise_irq\n", __FUNCTION__);
    qemu_set_irq(sc_link_pci_interrupt, 1);
}

void 
sc_low_pci_irq()
{
    PRINT_DEBUG("\n%s: low_irq\n", __FUNCTION__);
    qemu_set_irq(sc_link_pci_interrupt, 0);
}

static int
sc_link_pci_initfn(PCIDevice *pci_dev)
{
    SCLinkState *d = DO_UPCAST(SCLinkState, dev, pci_dev);
    uint8_t *pci_conf;
    int i;
    
    PRINT_DEBUG("%s: SystemC_Link init...\n", __FUNCTION__);

    pci_conf = d->dev.config;
    sc_link_configuration_write(pci_conf);
  
    d->mmio_index = cpu_register_io_memory(sc_link_pci_mem_read, sc_link_pci_mem_write, d);

    for(i=0; i < PCI_NUM_REGIONS -1; i++) {//-1 for the Exp ROM BAR
	    if(BAR_Regions[i][0] != 0)
	    { 
	        if(BAR_Regions[i][1] == PCI_BASE_ADDRESS_SPACE_IO) 
	        {
                //io region
		        PRINT_DEBUG("%s: Registering Bar %i as I/O BAR\n", __FUNCTION__, i);
		        pci_register_bar((PCIDevice *)d, i, BAR_Regions[i][0], PCI_BASE_ADDRESS_SPACE_IO, sc_link_pci_io_map);			
	        } else { 
                //mem region
		        PRINT_DEBUG("%s: Registering Bar %i as MEM BAR\n", __FUNCTION__, i);
		        pci_register_bar((PCIDevice *)d, i, BAR_Regions[i][0], PCI_BASE_ADDRESS_SPACE_MEMORY, sc_link_pci_mem_map);	
	        }
	    }
    }

    sc_link_pci_interrupt = d->dev.irq[0];
    PRINT_DEBUG("%s: Init done\n", __FUNCTION__);
    return 0;
}

void 
sc_link_pci_init(PCIBus *bus)
{
    pci_create_simple(bus, -1, "SCLink");
}

static PCIDeviceInfo sc_link_pci_info = {
    .qdev.name = "SCLink",
    .qdev.desc = "SystemC Link Device",
    .qdev.size = sizeof(SCLinkState),
    .init      = sc_link_pci_initfn,
};

static void 
sc_link_pci_register_devices(void)
{
    pci_qdev_register(&sc_link_pci_info);
}

device_init(sc_link_pci_register_devices)
